#include "compressedtextureimageprovider.h"

#include <QFile>
#include <QDebug>
#include <QtEndian>
#include <QScopedPointer>

#include <QOpenGLContext>
#include <QOpenGLFunctions>
#include <QOpenGLTexture>
#include <QQuickTextureFactory>
#include <QSGTexture>

#include <qopengl.h>

#ifndef PKM_ETC2_RGB_NO_MIPMAPS
#define PKM_ETC2_RGB_NO_MIPMAPS 0x01
#endif

#define PKM_HEADER_LENGTH 16
#define PKM_HEADER_PREAMBLE "PKM "
#define PKM_HEADER_PREAMBLE_LENGTH 4
#define PKM_HEADER_VERSION "20"
#define PKM_HEADER_VERSION_LENGTH 2

namespace {

// assuming (via the check below) that we contain PKM_ETC2_RGB_NO_MIPMAPS data.
struct PKMImage
{
    QSize effectiveSize;
    QSize originalSize;
    QByteArray data;
};

class CompressedSGTexture : public QSGTexture
{
    Q_OBJECT

public:
    CompressedSGTexture(const PKMImage &image);
    ~CompressedSGTexture();
    void bind() Q_DECL_OVERRIDE;
    bool hasAlphaChannel() const Q_DECL_OVERRIDE;
    bool hasMipmaps() const Q_DECL_OVERRIDE;

    int textureId() const Q_DECL_OVERRIDE;
    QSize textureSize() const Q_DECL_OVERRIDE;

private:
    PKMImage m_image;
    QScopedPointer<QOpenGLTexture> m_texture;
};

CompressedSGTexture::CompressedSGTexture(const PKMImage &image)
    : m_image(image)
    , m_texture(0)
{
}

CompressedSGTexture::~CompressedSGTexture()
{
    qDebug() << Q_FUNC_INFO << "Texture destroyed!";
}

void CompressedSGTexture::bind()
{
    if (!m_texture) {
        m_texture.reset(new QOpenGLTexture(QOpenGLTexture::Target2D));
        m_texture->setSize(m_image.effectiveSize.width(), m_image.effectiveSize.height());
        m_texture->setFormat(QOpenGLTexture::RGB8_ETC2);
        m_texture->setMipLevels(1);
        m_texture->setMipLevelRange(0, 0);
        m_texture->allocateStorage();
        m_texture->setCompressedData(m_image.data.size(), m_image.data.constData());
    }

    m_texture->bind();
}

bool CompressedSGTexture::hasAlphaChannel() const
{
    return false; // not supported, only PKM_ETC2_RGB_NO_MIPMAPS
}

bool CompressedSGTexture::hasMipmaps() const
{
    return false; // not supported, only PKM_ETC2_RGB_NO_MIPMAPS
}

int CompressedSGTexture::textureId() const
{
    return m_texture ? m_texture->textureId() : 0;
}

QSize CompressedSGTexture::textureSize() const
{
    return m_image.effectiveSize;
}


class CompressedTextureFactory : public QQuickTextureFactory
{

    Q_OBJECT
public:
    CompressedTextureFactory(const PKMImage &image);

    QSGTexture *createTexture(QQuickWindow *window) const Q_DECL_OVERRIDE;
    QImage image() const Q_DECL_OVERRIDE;
    int textureByteCount() const Q_DECL_OVERRIDE;
    QSize textureSize() const Q_DECL_OVERRIDE;

private:
    PKMImage m_image;
};

CompressedTextureFactory::CompressedTextureFactory(const PKMImage &image)
    : m_image(image)
{
}

QSGTexture *CompressedTextureFactory::createTexture(QQuickWindow *window) const
{
    Q_UNUSED(window);
    return new CompressedSGTexture(m_image);
}

QImage CompressedTextureFactory::image() const
{
    // FIXME/TODO: what's this method ever used for?
    // We can't easily get a QImage out of compressed texture data.
    return QImage();
}

int CompressedTextureFactory::textureByteCount() const
{
    return m_image.data.length();
}

QSize CompressedTextureFactory::textureSize() const
{
    return m_image.effectiveSize;
}

} // anonymous namespace


CompressedTextureImageProvider::CompressedTextureImageProvider()
    : QQuickImageProvider(QQmlImageProviderBase::Texture)
{
}

QQuickTextureFactory *CompressedTextureImageProvider::requestTexture(const QString &id, QSize *size, const QSize &requestedSize)
{
    Q_UNUSED(requestedSize);

    QFile imageFile(QStringLiteral(":/%1.pkm").arg(id));
    if (!imageFile.open(QFile::ReadOnly))
        return 0;

    QByteArray header = imageFile.read(PKM_HEADER_LENGTH);
    if (header.length() != PKM_HEADER_LENGTH) {
        qWarning() << "PKM header too short";
        return 0;
    }

    PKMImage image;
    const char *headerData = header.constData();

    // Parse the PKM header
    if (memcmp(headerData, PKM_HEADER_PREAMBLE, PKM_HEADER_PREAMBLE_LENGTH) != 0) {
        qWarning() << "Malformed PKM header (missing heading)";
        return 0;
    }
    headerData += 4;

    if (memcmp(headerData, PKM_HEADER_VERSION, PKM_HEADER_VERSION_LENGTH) != 0) {
        qWarning() << "Malformed PKM header (wrong version)";
        return 0;
    }
    headerData += 2;

#define UCC(x) (reinterpret_cast<const uchar *>(x))
    const quint16 dataType = qFromBigEndian<quint16>(UCC(headerData));
    if (dataType != PKM_ETC2_RGB_NO_MIPMAPS) {
        qWarning() << "Malformed PKM header (wrong data type)";
        return 0;
    }
    headerData += 2;

    image.effectiveSize.rwidth() = qFromBigEndian<quint16>(UCC(headerData));
    headerData += 2;
    image.effectiveSize.rheight() = qFromBigEndian<quint16>(UCC(headerData));
    headerData += 2;
    image.originalSize.rwidth() = qFromBigEndian<quint16>(UCC(headerData));
    headerData += 2;
    image.originalSize.rheight() = qFromBigEndian<quint16>(UCC(headerData));
    headerData += 2;
#undef UCC

    Q_ASSERT(image.effectiveSize.width() % 4 == 0);
    Q_ASSERT(image.effectiveSize.height() % 4 == 0);
    Q_ASSERT(headerData == header.constEnd());

    // Read out the payload
    const qint64 imageDataLength = ((image.effectiveSize.width() / 4) * (image.effectiveSize.height() / 4)) * 8;
    image.data = imageFile.read(imageDataLength);
    if (image.data.length() != imageDataLength) {
        qWarning() << "Malformed PKM file: payload too short";
        return 0;
    }

    if (!imageFile.atEnd()) {
        qWarning() << "Malformed PKM file: data after the payload";
        return 0;
    }

    if (size)
        *size = image.effectiveSize;

    return new CompressedTextureFactory(image);
}

#include "compressedtextureimageprovider.moc"

